/* GtkEditor - a source editor widget for GTK
 * Copyright (C) 1998, 2000 Thomas Mailund.
 *
 * The editor widget was written by Thomas Mailund, so bugs should be
 * reported to <mailund@daimi.au.dk>, not the gtk ppl.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <glib.h>
#include <gtk/gtk.h>

#include "internal.h"
#include "edit.h"
#include "highlight.h"

typedef struct _PropertyChange {
  int id;
  unsigned int offset;
  unsigned int length;
}PropertyChange;

/*
 * globals
 */
/* for recycling */
static GSList *g_free_text_props = NULL;

static void copy_gtk_text (char *buffer, GtkText *text, int pos, int length);

void
gtk_editor_install_text_properties (GtkEditor *editor, GList *props)
{
  GtkEditorTextProperty *prop;
  GList *cur;
  int i = 0;

  if (editor->text_properties)
    g_free (editor->text_properties);

  editor->n_text_properties = g_list_length (props);
  editor->text_properties = g_new (GtkEditorTextProperty, editor->n_text_properties);

  for (cur = props; cur != NULL; cur = cur->next) {
    prop = cur->data;
    editor->text_properties[i].font = gtk_editor_fontdup(prop->font);
    editor->text_properties[i].fore = gtk_editor_coldup(prop->fore);
    editor->text_properties[i].back = gtk_editor_coldup(prop->back);
    i++;
  }
}

GList *
gtk_editor_text_property_new (GdkFont *font, GdkColor *fore,
			      GdkColor *back, GList *next)
{
  GList *tmp = g_list_alloc ();
  GtkEditorTextProperty *new = g_new (GtkEditorTextProperty, 1);

  new->font = gtk_editor_fontdup(font);
  new->fore = gtk_editor_coldup(fore);
  new->back = gtk_editor_coldup(back);

  tmp->data = (gpointer)new;
  tmp->next = next;

  return tmp;
}

GSList *
_gtk_editor_add_property_change (GSList *text_changes,
				 int id, unsigned int offset,
				 unsigned int length)
{
  GSList *cur;
  PropertyChange *prop;

  if(text_changes
     && ((prop = text_changes->data) != NULL)
     && offset == prop->offset + prop->length
     && id == prop->id) {
    prop->length += length;
    return text_changes;
  }

  if (g_free_text_props) {
    cur = g_free_text_props;
    g_free_text_props = g_slist_remove_link (g_free_text_props, cur);
    prop = cur->data;
  } else {
    prop = g_new (PropertyChange, 1);
    cur = g_slist_alloc();
    cur->data = prop;
    cur->next = NULL;
  }

  prop->id = id;
  prop->offset = offset;
  prop->length = length;

  if(text_changes == NULL)
    return cur;
  return(g_slist_concat(cur, text_changes));
}

void
_gtk_editor_apply_property_changes (GtkEditor *editor,
				    char *text,
				    GSList *text_changes,
				    int offset)
{
  GSList *cur;
  PropertyChange *prop;
  int t_offset;

  if(text_changes == NULL)
    return;

  text_changes = g_slist_reverse(text_changes);

  for(cur = text_changes; cur != NULL;) {
    prop = cur->data;
    t_offset = (prop->offset >= g_insertion_point)?
               ((g_deletion_length > 0)? g_deletion_length : -g_insertion_length)
               : 0;
    gtk_text_set_point(GTK_TEXT(editor), prop->offset + t_offset);
    gtk_text_forward_delete (GTK_TEXT (editor), prop->length);
    gtk_text_insert(GTK_TEXT(editor), editor->text_properties[prop->id].font,
                    editor->text_properties[prop->id].fore,
                    editor->text_properties[prop->id].back,
                    (text + (prop->offset - offset)),
                    prop->length);
    text_changes = g_slist_remove_link(text_changes, cur);
    g_free_text_props = g_slist_concat(cur, g_free_text_props);
    cur = text_changes;
  }
}

void
_gtk_editor_insert_text_with_properties (GtkEditor *editor,
					 const char *text,
					 GSList *text_changes)
{
  GSList *cur;
  PropertyChange *prop;
  text_changes = g_slist_reverse(text_changes);
  for(cur = text_changes; cur != NULL;) {
    prop = cur->data;
    gtk_text_set_point(GTK_TEXT(editor), prop->offset);
    gtk_text_insert(GTK_TEXT(editor), editor->text_properties[prop->id].font,
                    editor->text_properties[prop->id].fore,
                    editor->text_properties[prop->id].back,
                    text + (prop->offset - g_insertion_point),
                    prop->length);
    text_changes = g_slist_remove_link(text_changes, cur);
    g_free_text_props = g_slist_concat(cur, g_free_text_props);
    cur = text_changes;
  }
}

#define MAX_BUFFER_LENGTH 1024

void
_gtk_editor_update_blocks (GtkEditor *editor,
			   GNode *base_block,
			   GNode *prev_block,
			   GNode *next_block,
			   GList *start,
			   int pos)
{
  GSList *cur;
  GList *cur_token;
  GNode *parent, *cur_block = base_block;
  GSList *prop_changes = NULL;
  PropertyChange *prop_change;
  Block *base = NULL, *block = NULL, *next = NULL, *temp_block;
  Token *token;
  int i, max_len = 0, block_id, prop_id, done = 0, changed = 1;
  char *buffer;
  static char sbuffer[MAX_BUFFER_LENGTH];

  if(base_block) {
    base = base_block->data;
    block = base;
  }
  if(next_block)
    next = next_block->data;

  for (cur_token = start; cur_token != NULL && !done; cur_token = cur_token->next) {
    token = cur_token->data;
    /* When we reach the end of the base block we are highlighting we are done. */

    if (cur_block && cur_token->prev == block->end) {
      prev_block = cur_block;
      next_block = cur_block->next;
      cur_block = cur_block->parent;
      if(cur_block)
        block = cur_block->data;
      if(next_block)
        next = next_block->data;
    }

    if (cur_block) {
      if (editor->token_entries[token->id].ends) {
        for(parent = cur_block; parent != NULL; parent = parent->parent) {
          temp_block = parent->data;
          for (i = 0; i < editor->token_entries[token->id].n_ends; i++) {
            if (temp_block->id == editor->token_entries[token->id].ends[i]) {
              temp_block->end = cur_token;
              if (next_block && cur_token == next->begin) {
                parent = next_block;
                if(next_block->children)
                  next_block = next_block->children;
                else
                  next_block = next_block->next;
                _gtk_editor_remove_block (editor, parent);
              }
              goto out;
            }
          }
        }
      }
    } else if (((block_id = editor->token_entries[token->id].begin) != -1)
               && ((cur_block && editor->block_entries[block->id].nest
                    && editor->block_entries[block_id].precedence >=
                       editor->block_entries[block->id].precedence)
                   || !cur_block)) {
      if (next_block && cur_token == next->begin) {
        cur_block = next_block;
        block = next;
        prev_block = NULL;
        next_block = next_block->children;
        if(next_block)
          next = next_block->data;
        goto out;
      }
      if (cur_block)
        parent = cur_block->parent;
      else
        parent = NULL;
      cur_block = _gtk_editor_get_new_block();
      if (parent && parent->children == NULL)
        parent->children = cur_block;
      if (!editor->blocks)
        editor->blocks = cur_block;
      if (prev_block)
        prev_block->next = cur_block;
      if (next_block)
        next_block->prev = cur_block;
      cur_block->next = next_block;
      cur_block->prev = prev_block;
      cur_block->parent = parent;
      cur_block->children = NULL;
      block = cur_block->data;
      block->begin = cur_token;
      block->end = NULL;
      block->id = block_id;
      if (next_block)
        next = next_block->data;
    }
    out:

    if (cur_block)
      prop_id = editor->block_entries[block->id].prop_id;
    else
      prop_id = editor->token_entries[token->id].prop_id;

    if (token->prop_id != prop_id) {
      token->prop_id = prop_id;
      prop_changes = _gtk_editor_add_property_change (prop_changes,
						      prop_id, pos,
						      token->length);
      changed = 1;
    }
    else
      changed = 0;
    pos += token->length;
  }

  for (cur = prop_changes; cur != NULL; cur = cur->next) {
    prop_change = cur->data;
    if (prop_change->length > max_len)
      max_len = prop_change->length;
  }

  if (max_len > MAX_BUFFER_LENGTH)
    buffer = g_new(char, max_len);
  else
    buffer = sbuffer;

  gtk_text_freeze(GTK_TEXT(editor));
  for (cur = prop_changes; cur != NULL;) {
    prop_change = cur->data;
    copy_gtk_text (buffer, GTK_TEXT(editor), prop_change->offset, prop_change->length);
    gtk_text_set_point(GTK_TEXT(editor), prop_change->offset);
    gtk_text_forward_delete (GTK_TEXT (editor), prop_change->length);
    gtk_text_insert(GTK_TEXT(editor), editor->text_properties[prop_change->id].font,
                    editor->text_properties[prop_change->id].fore,
                    editor->text_properties[prop_change->id].back,
                    buffer, prop_change->length);
    prop_changes = g_slist_remove_link(prop_changes, cur);
    g_free_text_props = g_slist_concat(cur, g_free_text_props);
    cur = prop_changes;
  }
  gtk_text_thaw(GTK_TEXT(editor));

  if (buffer != sbuffer)
    g_free (buffer);
}

static void
copy_gtk_text (char *buffer, GtkText *text, int pos, int length)
{
  if (pos < text->gap_position) {
    if (pos + length < text->gap_position) {
      /* before gap */
      memcpy(buffer, text->text.ch + pos, length);

    } else {
      /* before and intersecting gap */
      int l = text->gap_position - pos;
      memcpy (buffer, text->text.ch + pos, l);
      memcpy (buffer,
              text->text.ch + text->gap_position + text->gap_size,
              length - l);
    }
  } else {
    /* starts inside or after gap */
    memcpy(buffer, text->text.ch + text->gap_size + pos, length);
  }
}

